import logging
import os
import sys
from importlib import import_module
from typing import TYPE_CHECKING

import yaml
from django.conf import settings
from kitchenai.contrib.kitchenai_sdk.kitchenai import KitchenAIApp
from kitchenai.core.models import KitchenAIManagement

if TYPE_CHECKING:
    from ninja import NinjaAPI

logger = logging.getLogger("kitchenai.core.utils")

def load_config_from_db():
    config = {}
    mgmt = KitchenAIManagement.objects.get(name="kitchenai_management")

    app = mgmt.kitchenaimodules_set.filter(is_root=True).first()
    config["app"] = yaml.safe_load(app.name)
    return config

def update_installed_apps(self, apps):
    if apps:
        settings.INSTALLED_APPS += tuple(apps)
        self.stdout.write(self.style.SUCCESS(f'Updated INSTALLED_APPS: {settings.INSTALLED_APPS}'))

def import_modules(module_paths):
    for name, path in module_paths.items():
        try:
            module_path, instance_name = path.split(':')
            module = import_module(module_path)
            instance = getattr(module, instance_name)
            globals()[name] = instance
            logger.info(f'Imported {instance_name} from {module_path}')
        except (ImportError, AttributeError) as e:
            logger.error(f"Error loading module '{path}': {e}")

def import_cookbook(module_path):
    try:
        module_path, instance_name = module_path.split(':')
        module = import_module(module_path)
        instance = getattr(module, instance_name)
        print(f'Imported {instance_name} from {module_path}')
        return instance
    except (ImportError, AttributeError) as e:
        print(f"Error loading module '{e}")



def setup(api: "NinjaAPI"):
    # # Load configuration from the database
    config = load_config_from_db()

    if not config:
        logger.error('No configuration found. Please run "kitchenai init" first.')
        return

    # Update INSTALLED_APPS and import modules
    # self.update_installed_apps(config.get('installed_apps', []))

    # self.import_modules(config.get('module_paths', {}))

    # Determine the user's project root directory (assumes the command is run from the user's project root)
    project_root = os.getcwd()

    # Add the user's project root directory to the Python path
    if project_root not in sys.path:
        sys.path.insert(0, project_root)

    #importing main app
    try:
        print(config["app"])

        module_path, instance_name = config["app"].split(':')
        module = import_module(module_path)
        instance = getattr(module, instance_name)

        logger.info(f'Imported {instance_name} from {module_path}')
        if isinstance(instance, KitchenAIApp):
            logger.info(f'{instance_name} is a valid KitchenAIApp instance.')
            api.add_router(f"/{instance._namespace}", instance._router)
        else:
            logger.error(f'{instance_name} is not a valid KitchenAIApp instance.')

    except (ImportError, AttributeError) as e:
        logger.error(f"Error loading module: {e}")



import logging

import pytest
from django.test.utils import override_settings

pytest_plugins = []  # type: ignore


def pytest_configure(config):
    logging.disable(logging.CRITICAL)


TEST_SETTINGS = {
    "DEBUG": False,
    "CACHES": {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    },
    "EMAIL_BACKEND": "django.core.mail.backends.locmem.EmailBackend",
    "LOGGING_CONFIG": None,
    "PASSWORD_HASHERS": [
        "django.contrib.auth.hashers.MD5PasswordHasher",
    ],
    "Q_CLUSTER": {
        "sync": True,
    },
    "STORAGES": {
        "default": {
            "BACKEND": "django.core.files.storage.InMemoryStorage",
        },
        "staticfiles": {
            "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
        },
    },
    "WHITENOISE_AUTOREFRESH": True,
}


@pytest.fixture(autouse=True, scope="session")
def use_test_settings():
    with override_settings(**TEST_SETTINGS):
        yield



from django.db import models
from falco_toolbox.models import TimeStamped


class KitchenAIManagement(TimeStamped):
    name = models.CharField(max_length=255, primary_key=True, default="kitchenai_management")
    project_name = models.CharField(max_length=255)
    version = models.CharField(max_length=255)
    description = models.TextField(default="")

    def __str__(self):
        return self.name


class KitchenAIModules(TimeStamped):
    name = models.CharField(max_length=255, primary_key=True)
    kitchen = models.ForeignKey(KitchenAIManagement, on_delete=models.CASCADE)
    is_root = models.BooleanField(default=False)

    def __str__(self):
        return self.name


class KitchenAIPlugins(TimeStamped):
    name = models.CharField(max_length=255, primary_key=True)
    kitchen = models.ForeignKey(KitchenAIManagement, on_delete=models.CASCADE)

    def __str__(self):
        return self.name


class KitchenAIDependencies(TimeStamped):
    name = models.CharField(max_length=255, primary_key=True)
    kitchen = models.ForeignKey(KitchenAIManagement, on_delete=models.CASCADE)

    def __str__(self):
        return self.name



import logging
import sys

import django
import typer
from django.conf import settings
from cookiecutter.main import cookiecutter

logger = logging.getLogger(__name__)
app = typer.Typer()

@app.command()
def add(module: str = typer.Argument("app.kitchen:kitchen")):
    from django.core.management import execute_from_command_line

    execute_from_command_line(["manage", "add_module", module])

@app.command()
def init():
    from django.core.management import execute_from_command_line
    execute_from_command_line(["manage", "migrate"])

    execute_from_command_line(["manage", "init"])


@app.command()
def qcluster() -> None:
    """Run Django-q cluster."""
    from django.core.management import execute_from_command_line

    # execute_from_command_line(["manage", "qcluster", *argv[2:]])
    execute_from_command_line(["manage", "qcluster"])


@app.command()
def runserver() -> None:
    """Run Django runserver."""
    # from django.core.management import execute_from_command_line
    #sys.argv pop to remove the command line "dev" before extending the gunicorn command
    sys.argv.pop(1)
    django.setup()
    from kitchenai.api import api
    from kitchenai.core.utils import setup

    setup(
        api
    )
    _run_dev_uvicorn(sys.argv)

@app.command()
def run() -> None:
    """Run Django runserver."""
    sys.argv.pop(1)
    django.setup()
    from kitchenai.api import api
    from kitchenai.core.utils import setup

    setup(
        api
    )

    _run_uvicorn(sys.argv)


@app.command()
def dev(address: str ="0.0.0.0:8000"):
    """
    Reads the kitchen config file, reads the application file and runs the KitchenAI server
    """
    commands = {"server": "kitchenai runserver"}
    if "django_tailwind_cli" in settings.INSTALLED_APPS:
        commands["tailwind"] = "django-admin tailwind watch"
    if "tailwind" in settings.INSTALLED_APPS:
        commands["tailwind"] = "django-admin tailwind start"
    if "django_q" in settings.INSTALLED_APPS:
        commands["qcluster"] = "kitchenai qcluster"

    typer.echo(f"[INFO] starting development server on {address}")

    # call_command("migrate")
    _run_with_honcho(commands)

@app.command()
def manage() -> None:
    """Run Django's manage command."""
    from django.core.management import execute_from_command_line

    # execute_from_command_line(argv[1:])
    execute_from_command_line(["manage"])

@app.command()
def setup():
    """Run some project setup tasks"""
    from django.core.management import execute_from_command_line
    import os

    execute_from_command_line(["manage", "migrate"])
    execute_from_command_line(["manage", "setup_periodic_tasks"])

    # Set environment variables for superuser credentials
    email = os.environ.get("DJANGO_SUPERUSER_EMAIL", "admin@localhost")
    password = os.environ.get("DJANGO_SUPERUSER_PASSWORD", "admin")
    username = os.environ.get("DJANGO_SUPERUSER_USERNAME", email.split("@")[0])

    if password == "admin":
        #set it 
        os.environ["DJANGO_SUPERUSER_PASSWORD"] = "admin"
    execute_from_command_line(
        ["manage", "createsuperuser", "--noinput", "--traceback", "--email", email, "--username", username]
    )

@app.command()
def build():
    """
    Reads the kitchen config file, reads the application file and runs the KitchenAI server
    """
    from django.core.management import execute_from_command_line
    django.setup()

    execute_from_command_line(["manage", "build_container"])


@app.command()
def new():
    """
    Reads the kitchen config file, reads the application file and runs the KitchenAI server
    """

    cookiecutter("https://github.com/epuerta9/cookiecutter-cookbook.git", output_dir=".")



def _run_with_honcho(commands: dict):
    from honcho.manager import Manager

    manager = Manager()
    for name, cmd in commands.items():
        manager.add_process(name, cmd)
    try:
        manager.loop()
    finally:
        manager.terminate()



def _run_uvicorn(argv: list) -> None:
    """
    Run gunicorn + uvicorn workers server.
    https://docs.gunicorn.org/en/stable/settings.html
    https://adamj.eu/tech/2021/12/29/set-up-a-gunicorn-configuration-file-and-test-it/
    """

    import multiprocessing
    from gunicorn.app import wsgiapp  # for gunicorn

    workers = multiprocessing.cpu_count() * 2 + 1
    gunicorn_args = [
        "kitchenai.asgi:application",  # Replace WSGI with ASGI app
        "--bind",
        "0.0.0.0:8000",
        # "unix:/run/kitchenai_demo.gunicorn.sock",  # Use this if you're using a socket file
        "--max-requests",
        "1000",
        "--max-requests-jitter",
        "50",
        "--workers",
        str(workers),
        "--worker-class",
        "uvicorn.workers.UvicornWorker",  # Use Uvicorn worker for ASGI
        "--access-logfile",
        "-",
        "--error-logfile",
        "-",
    ]
    argv.extend(gunicorn_args)

    wsgiapp.run()


def _run_dev_uvicorn(argv: list) -> None:
    """
    Run gunicorn + uvicorn workers server.
    https://docs.gunicorn.org/en/stable/settings.html
    https://adamj.eu/tech/2021/12/29/set-up-a-gunicorn-configuration-file-and-test-it/
    """
    from gunicorn.app import wsgiapp  # for gunicorn

    workers = 2
    gunicorn_args = [
        "kitchenai.asgi:application",  # Replace WSGI with ASGI app
        "--bind",
        "0.0.0.0:8000",
        # "unix:/run/kitchenai_demo.gunicorn.sock",  # Use this if you're using a socket file
        "--max-requests",
        "1000",
        "--max-requests-jitter",
        "50",
        "--workers",
        str(workers),
        "--worker-class",
        "uvicorn.workers.UvicornWorker",  # Use Uvicorn worker for ASGI
        "--access-logfile",
        "-",
        "--error-logfile",
        "-",
    ]
    argv.extend(gunicorn_args)

    wsgiapp.run()



import asyncio
import functools
import logging

from ninja import Router

# Set up logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

class KitchenAIApp:
    def __init__(self, router: Router = None, namespace: str = 'default'):
        self._namespace = namespace
        self._router = router if router else Router()

    # Helper method to create a decorator for a given route type
    def _create_decorator(self, route_type: str, method: str, label: str):
        def decorator(func, **route_kwargs):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                if asyncio.iscoroutinefunction(func):
                    return await func(*args, **kwargs)
                else:
                    loop = asyncio.get_event_loop()
                return await loop.run_in_executor(None, lambda: func(*args, **kwargs))

            # Define the path for the route using the namespace and label
            route_path = f"/{route_type}/{label}"

            # Register the route using add_api_operation
            self._router.add_api_operation(
                path=route_path,
                methods=[method],  # Customize as needed (GET, PUT, etc.)
                view_func=wrapper,
                **route_kwargs  # Pass the custom kwargs defined by the user in the decorator
            )
            logger.debug(f"Registered route: {route_path}")
            return wrapper
        return decorator

    # Query decorator
    def query(self, label: str, **route_kwargs):
        return self._create_decorator('query',"POST", label)

    # Storage decorator
    def storage(self, label: str, **route_kwargs):
        return self._create_decorator('storage', "POST", label)

    # Embedding decorator
    def embedding(self, label: str, **route_kwargs):
        return self._create_decorator('embedding', "POST", label)

    # Runnable decorator (for chaining multiple tasks)
    def runnable(self, label: str, **route_kwargs):
        return self._create_decorator('runnable', "POST", label)



from django.contrib import admin

from .models import KitchenAIManagement
from .models import KitchenAIModules

@admin.register(KitchenAIManagement)
class KitchenAIAdmin(admin.ModelAdmin):
    pass



@admin.register(KitchenAIModules)
class KitchenAIModuleAdmin(admin.ModelAdmin):
    pass
